-- Addon & Module System
-- Author: kurapica.igas@gmail.com
-- Create Date : 2011/02/26
-- ChangeLog   :
--						2011/03/16	Modify the enable-disable system.

----------------------------------------------------------------------------------------------------------------------------------------
--- Addon is an class type for addons in IGAS.The addon class supply some methods, to help authors to manager their addons easily.
-- an addon is also a private scope using to contains it's own functions and variables.
-- addons can access _G's functions and variables, and making an image in itself at the same time.When it access a frame in _G, it'll automatically convert it to an IGAS' frame into the addon' scope.
-- @name Addon
-- @class table
-- @field _Addon the addon self, readonly
-- @field _Name the addon's name, readonly
-- @field _Version the addon's version
-- @field _Enabled whether the addon is enabled or disabled.
-- @field _AutoWrapper whether the addon would auto making wrappers for frames defined in _G when the addon access it.
-- @field _Metadata the addon's metatable
----------------------------------------------------------------------------------------------------------------------------------------

----------------------------------------------------------------------------------------------------------------------------------------
--- Module is an class type for module in IGAS.a module is contained in an addon or another module.
-- a module is also a private scope.It can access all resources defined in it's container.
-- @name Module
-- @class table
-- @field _M the module self, readonly
-- @field _Name the module's name, readonly
-- @field _Version the module's version
-- @field _Enabled whether the module is enabled or disabled.
-- @field _Parent the module's parent module(addon)
----------------------------------------------------------------------------------------------------------------------------------------

------------------------------------------------------
-- Version Check & Environment
------------------------------------------------------
do
	local version = 2

	local _Meta = getmetatable(IGAS)

	-- Version Check
	if _Meta._IGAS_Addon_Version and _Meta._IGAS_Addon_Version >= version then
		return
	end

	_Meta._IGAS_Addon_Version = version

	-- Class Environment
	_Meta._Addon_ENV = _Meta._Addon_ENV or {}

	setmetatable(_Meta._Addon_ENV, {
		__index = function(self,  key)
			if type(key) == "string" and key ~= "_G" and key:find("^_") then
				return
			end

			if _Meta._Class_KeyWords[key] then
				return _Meta._Class_KeyWords[key]
			end

			if _G[key] then
				rawset(self, key, _G[key])
				return rawget(self, key)
			end
		end,
	})

	-- Special
	_Meta._Addon_ENV._Class_KeyWords = _Meta._Class_KeyWords

	setfenv(1, _Meta._Addon_ENV)

	-- Scripts
	strtrim = strtrim or function(s)
	  return (s:gsub("^%s*(.-)%s*$", "%1")) or ""
	end

	wipe = wipe or function(t)
		for k in pairs(t) do
			t[k] = nil
		end
	end

	errorhandler = errorhandler or function(err)
		return geterrorhandler and geterrorhandler()(err) or print(err)
	end

	tinsert = tinsert or table.insert
	tremove = tremove or table.remove
end

namespace "System"

IGAS = IGAS

------------------------------------------------------
-- Storage Definition
------------------------------------------------------
_Addon = _Addon or {}

------------------------------------------------------
-- Addon
------------------------------------------------------
class "Addon"
	inherit "Object"

	_MetaWK = _MetaWK or {__mode = "k"}
	_MetaWV = _MetaWV or {__mode = "v"}
	_MetaWKV = _MetaWKV or {__mode = "kv"}
	_Class_KeyWords = _Class_KeyWords

	local type = type
	local strtrim = strtrim
	local rawset = rawset
	local rawget = rawget
	local wipe = wipe
	local setmetatable = setmetatable
	local errorhandler = errorhandler
	local tinsert = tinsert
	local tremove = tremove
	local next = next
	local pairs = pairs
	local pcall = pcall
	local error = error
	local getfenv = getfenv
	local setfenv = setfenv
	local _G = _G
	local CreateFrame = CreateFrame
	local GetAddOnMetadata = GetAddOnMetadata
	local IGAS = IGAS

	------------------------------------------------------
	-- Storage Definition
	------------------------------------------------------
	do
		_Addon = _Addon
		_Info = _Info or setmetatable({}, _MetaWK)

		_Logined = _Logined or false
	end

	------------------------------------------------------
	-- _EventManager
	--
	-- _EventManager:Register(event, obj)
	-- _EventManager:IsRegistered(event, obj)
	-- _EventManager:Unregister(event, obj)
	-- _EventManager:UnregisterAll(obj)
	------------------------------------------------------
	do
		------------------------------------------------------
		-- For World of Warcraft
		------------------------------------------------------
		if CreateFrame then
			_EventManager = _EventManager or CreateFrame("Frame")
			_EventDistribution =  _EventDistribution or {}

			_UsedEvent = _UsedEvent or {}

			-- Register
			function _EventManager:Register(event, obj)
				if type(event) == "string" and event ~= "" then
					self:RegisterEvent(event)

					if self:IsEventRegistered(event) then
						_EventDistribution[event] = _EventDistribution[event] or setmetatable({}, _MetaWK)

						_EventDistribution[event][obj] = true
					else
						return Object.RegisterEvent(obj, event)
					end
				end
			end

			-- IsRegistered
			function _EventManager:IsRegistered(event, obj)
				if _EventDistribution[event] then
					return _EventDistribution[event][obj] or false
				else
					return Object.IsEventRegistered(obj, event)
				end
			end

			-- Unregister
			function _EventManager:Unregister(event, obj)
				if type(event) == "string" and event ~= "" then
					if _EventDistribution[event] then
						_EventDistribution[event][obj] = nil

						if not next(_EventDistribution[event]) and not _UsedEvent[event] then
							self:UnregisterEvent(event)
						end
					else
						return Object.UnregisterEvent(obj, event)
					end
				end
			end

			-- UnregisterAll
			function _EventManager:UnregisterAll(obj)
				for event, data in pairs(_EventDistribution) do
					if data[obj] then
						data[obj] = nil

						if not next(data) and not _UsedEvent[event] then
							self:UnregisterEvent(event)
						end
					end
				end

				return Object.UnregisterAllEvents(obj)
			end

			--  Special Settings for _EventManager
			do
				function _EventManager:OnEvent(event, ...)
					local chk, ret

					if type(self[event]) == "function" then
						chk, ret = pcall(self[event], self, ...)
						if not chk then
							errorhandler(ret)
						end
					end

					if _EventDistribution[event] then
						for obj in pairs(_EventDistribution[event]) do
							Object.Fire(obj, "OnEvent", event, ...)
						end
					end
				end

				-- Loading
				local function loading(self)
					Object.Fire(self, "OnLoad")

					if _Info[self].Module then
						for _, mdl in pairs(_Info[self].Module) do
							loading(mdl)
						end
					end
				end

				-- Enabling
				local function enabling(self)
					if not _Info[self].Disabled then
						Object.Fire(self, "OnEnable")

						if _Info[self].Module then
							for _, mdl in pairs(_Info[self].Module) do
								enabling(mdl)
							end
						end
					end
				end

				function _EventManager:ADDON_LOADED(addonName)
					local addon = _Addon[addonName]

					if addon and _Info[addon].SavedVariables then
						-- relocate savedvariables
						for svn in pairs(_Info[addon].SavedVariables) do
							_G[svn] = _G[svn] or {}
							addon[svn] = _G[svn]
						end

						_Info[addon].SavedVariables = nil
					end

					if addon then
						-- Fire OnLoad, addon no need register ADDON_LOADED self.
						loading(addon)

						if _Logined then
							-- enable addon if player was already in game.
							enabling(addon)
						end
					end
				end

				function _EventManager:PLAYER_LOGIN()
					for _, addon in pairs(_Addon) do
						if _Info[addon].SavedVariables then
							-- relocate savedvariables
							for svn in pairs(_Info[addon].SavedVariables) do
								_G[svn] = _G[svn] or {}
								addon[svn] = _G[svn]
							end

							_Info[addon].SavedVariables = nil
						end

						-- Enabling addons
						enabling(addon)
					end

					_Logined = true
				end

				_UsedEvent.ADDON_LOADED = true
				_UsedEvent.PLAYER_LOGIN = true

				_EventManager:Hide()
				_EventManager:UnregisterAllEvents()
				for event, flag in pairs(_UsedEvent) do
					if flag then
						_EventManager:RegisterEvent(event)
					end
				end

				_EventManager:SetScript("OnEvent", _EventManager.OnEvent)
			end
		end
	end

	------------------------------------------------------
	-- Module
	------------------------------------------------------
	class "Module"
		inherit "Object"

		_MetaWK = _MetaWK
		_MetaWV = _MetaWV
		_MetaWKV = _MetaWKV
		_Class_KeyWords = _Class_KeyWords

		local type = type
		local strtrim = strtrim
		local rawset = rawset
		local rawget = rawget
		local wipe = wipe
		local setmetatable = setmetatable
		local errorhandler = errorhandler
		local tinsert = tinsert
		local tremove = tremove
		local next = next
		local pairs = pairs
		local pcall = pcall
		local error = error
		local getfenv = getfenv
		local setfenv = setfenv
		local _G = _G
		local CreateFrame = CreateFrame
		local GetAddOnMetadata = GetAddOnMetadata

		------------------------------------------------------
		-- Storage Definition
		------------------------------------------------------
		do
			_Info = _Info
		end

		------------------------------------------------------
		-- Script
		------------------------------------------------------
		do
			script "OnLoad"
			script "OnEnable"
			script "OnDisable"
			script "OnDispose"
			script "OnSlashCmd"
		end

		------------------------------------------------------
		-- Method
		------------------------------------------------------
		do
			if _EventManager then
				------------------------------------
				--- Register a event
				-- @name Module:RegisterEvent
				-- @class function
				-- @param event the event's name
				-- @return nil
				-- @Usage Module:RegisterEvent("CUSTOM_EVENT_1")
				------------------------------------
				function RegisterEvent(self, event)
					if type(event) == "string" and event ~= "" then
						return _EventManager:Register(event, self)
					else
						error(("Usage : Module:RegisterEvent(event) : 'event' - string expected, got %s."):format(type(event) == "string" and "empty string" or type(event)), 2)
					end
				end

				------------------------------------
				--- Check if  a event is registered for an Module
				-- @name Module:IsEventRegistered
				-- @class function
				-- @param event the event's name
				-- @return true if the event is registered to that Module
				-- @Usage Module:IsEventRegistered("CUSTOM_EVENT_1")
				------------------------------------
				function IsEventRegistered(self, event)
					if type(event) == "string" and event ~= "" then
						return _EventManager:IsRegistered(event, self)
					else
						error(("Usage : Module:IsRegistered(event) : 'event' - string expected, got %s."):format(type(event) == "string" and "empty string" or type(event)), 2)
					end
				end

				------------------------------------
				--- Undo register all events for the Module
				-- @name Module:UnregisterAllEvents
				-- @class function
				-- @return nil
				-- @Usage Module:UnregisterAllEvents()
				------------------------------------
				function UnregisterAllEvents(self)
					return _EventManager:UnregisterAll(self)
				end

				------------------------------------
				--- Undo Register a event for an Module
				-- @name Module:UnregisterEvent
				-- @class function
				-- @param event the event's name
				-- @return nil
				-- @Usage Module:UnregisterEvent("CUSTOM_EVENT_1")
				------------------------------------
				function UnregisterEvent(self, event)
					if type(event) == "string" and event ~= "" then
						return _EventManager:Unregister(event, self)
					else
						error(("Usage : Module:Unregister(event) : 'event' - string expected, got %s."):format(type(event) == "string" and "empty string" or type(event)), 2)
					end
				end
			end

			------------------------------------
			--- Enable the Module.
			-- @name Module:Enable
			-- @class function
			-- @return nil
			-- @usage Module:Enable()
			------------------------------------
			function Enable(self)
				if _Info[self].Disabled then
					if _Info[_Info[self].Parent].Disabled then
						error("The module's parent module(addon) is disabled, can't enable it.", 2)
					end

					_Info[self].Disabled= nil
					_Info[self].DefaultState = nil

					if _Logined then Object.Fire(self, "OnEnable") end

					if not _Info[self].Module then return end

					for _, mdl in pairs(_Info[self].Module) do
						if _Info[mdl].DefaultState ~= false then
							Enable(mdl)
						end
					end
				end
			end

			------------------------------------
			--- Disable the Module.The Module won't receive events.
			-- @name Module:Disable
			-- @class function
			-- @return nil
			-- @usage Module:Disable()
			------------------------------------
			function Disable(self)
				if not _Info[self].Disabled then
					_Info[self].Disabled = true

					if _Logined then Object.Fire(self, "OnDisable") end

					if not _Info[self].Module then return end

					for _, mdl in pairs(_Info[self].Module) do
						_Info[mdl].DefaultState = not _Info[mdl].Disabled
						if not _Info[mdl].Disabled then
							Disable(mdl)
						end
					end
				else
					_Info[self].DefaultState = false
				end
			end

			------------------------------------
			--- Check if the Module is enabled
			-- @name Module:IsEnabled
			-- @class function
			-- @return true if the Module is enabled
			-- @usage Module:IsEnabled()
			------------------------------------
			function IsEnabled(self)
				return not _Info[self].Disabled
			end

			------------------------------------
			--- Add a frame to Module.
			-- @name Module:AddFrame
			-- @class function
			-- @param frame the user-defined frame that needed to be add to the Module, must be the parent of other user-defined frames, will be auto-diposed when Module is dispoing.Also make the frame can be access in the addon's scope by it's name.
			-- @return nil
			-- @usage Module:AddFrame(MyFrame1)
			------------------------------------
			function AddFrame(self, frame)
				if type(frame) == "table" and frame.Name then
					_Info[self].Frames = _Info[self].Frames or setmetatable({}, _MetaWK)
					_Info[self].Frames[frame] = true
					self[frame.Name] = frame
				end
			end

			------------------------------------
			--- Remove a frame from Module.
			-- @name Module:RemoveFrame
			-- @class function
			-- @param frame the user-defined frame that needed to be removed from the Module
			-- @return nil
			-- @usage Module:RemoveFrame(MyFrame1)
			------------------------------------
			function RemoveFrame(self, frame)
				if frame and _Info[self].Frames[frame] then
					_Info[self].Frames[frame] = nil
					if self[frame.Name] == frame then
						self[frame.Name] = nil
					end
				end
			end

			------------------------------------
			--- This is using to dispose a module, and set the enviroment to it's container if needed.
			-- @name Module:Dispose
			-- @class function
			-- @return nil
			-- @usage Module:Dispose()
			------------------------------------
			function Dispose(self)
				local info = _Info[self]
				local chk, ret

				Object.Fire(self, "OnDispose")

				if info.Frames then
					for fr in pairs(info.Frames) do
						if type(fr) == "table" and type(fr.Dispose) == "function" then
							chk, ret = pcall(fr.Dispose, fr)

							if not chk then
								errorhandler(ret)
							end
						end
					end
				end

				UnregisterAllEvents(self)

				if info.Module then
					for _, mdl in pairs(info.Module) do
						chk, ret = pcall(Dispose, mdl)

						if not chk then
							errorhandler(ret)
						end
					end
				end

				if info.Parent then
					if _Info[info.Parent].Module then
						_Info[info.Parent].Module[info.Name] = nil
					end
					if info.Parent[info.Name] == self then
						info.Parent[info.Name] = nil
					end
				end

				if getfenv(2) == self then
					if info.Parent then
						setfenv(2, info.Parent)
					else
						setfenv(2, _G)
					end
				end

				wipe(info)
				_Info[self] = nil

				return Object.Dispose(self)
			end

			------------------------------------
			--- Create or get a child-module of this module.And set the environment to the child-module
			-- @name Module:NewModule
			-- @class function
			-- @param name the child-module's name,can be "Mdl1.SubMdl1.SubSubMdl1", use dot to concat
			-- @param version Optional,used for version control, if there is another version, and the old version is equal or great than this one, return nil, else return the child-module
			-- @return the child-module
			-- @usage Module:NewModule("Mdl_MyFrame1")
			------------------------------------
			function NewModule(self, name, version)
				if type(name) ~= "string" or strtrim(name) == "" then
					error(("Usage : Module:NewModule(name, [version]) : 'name' - string expected, got %s."):format(type(name) == "string" and "empty string" or type(name)), 2)
				end

				if version ~= nil then
					if type(version) ~= "number" then
						error(("Usage : Module:NewModule(name, [version]) : 'version' - number expected, got %s."):format(type(version)), 2)
					elseif version < 1 then
						error("Usage : Module:NewModule(name, [version]) : 'version' - must be greater than 0.", 2)
					end
				end

				local module = self

				for sub in name:gmatch("[^%.]+") do
					sub = sub and strtrim(sub)
					if not sub or sub =="" then return end

					module = Module(module, sub)

					if not module then return end
				end

				if not module or module == self then return end

				if version and _Info[module].Version and _Info[module].Version >= version then
					return
				elseif version then
					_Info[module].Version = version
				end

				setfenv(2, module)

				return module
			end

			------------------------------------
			--- Get the child-module of this module.
			-- @name Module:GetModule
			-- @class function
			-- @param name the child-module's name,can be "Mdl1.SubMdl1.SubSubMdl1", use dot to concat
			-- @param create if true, the module will be created if not existed
			-- @return the child-module
			-- @usage Module:GetModule("Mdl_MyFrame1")
			------------------------------------
			function GetModule(self, name, create)
				if type(name) ~= "string" or strtrim(name) == "" then
					error(("Usage : Module:GetModule(name, [create]) : 'name' - string expected, got %s."):format(type(name) == "string" and "empty string" or type(name)), 2)
				end

				local module = self

				if create then
					for sub in name:gmatch("[^%.]+") do
						sub = sub and strtrim(sub)
						if not sub or sub =="" then return end

						module = Module(module, sub)

						if not module then return end
					end
				else
					for sub in name:gmatch("[^%.]+") do
						sub = sub and strtrim(sub)
						if not sub or sub =="" then return end

						module =  _Info[module].Module and _Info[module].Module[sub]

						if not module then return end
					end
				end

				if module == self then return end

				return module
			end
		end

		------------------------------------------------------
		-- Special keywords
		------------------------------------------------------
		_Special_KeyWord = _Special_KeyWord or {}
		do
			------------------------------------
			--- import classes from the given name's namespace to the current Addon
			-- @name import
			-- @class function
			-- @param name the namespace's name list, using "." to split.
			-- @usage import "System.Widget"
			------------------------------------
			function _Special_KeyWord.import(name)
				if type(name) ~= "string" then
					error([[Usage: import "namespaceA.namespaceB"]], 2)
				end

				if type(name) == "string" and name:find("%.%s*%.") then
					error("the namespace 's name can't have empty string between dots.", 2)
				end

				local env = getfenv(2)

				local info = _Info[env]

				if not info then
					error("can't use import here.", 2)
				end

				local ns = IGAS:GetNameSpace(name)

				if not ns then
					error(("no namespace is found with name : %s"):format(name), 2)
				end

				info.Import = info.Import or {}

				for _, v in ipairs(info.Import) do
					if v == ns then
						return
					end
				end

				tinsert(info.Import, ns)
			end
		end

		------------------------------------------------------
		-- Property
		------------------------------------------------------
		do
			-- _M
			property "_M" {
				Get = function(self)
					return self
				end,
			}

			-- _Name
			property "_Name" {
				Get = function(self)
					return _Info[self].Name
				end,
			}

			-- _Version
			property "_Version" {
				Get = function(self)
					return _Info[self].Version or 0
				end,
				Set = function(self, version)
					if version > (_Info[self].Version or 0) then
						_Info[self].Version = version
					else
						error(("The version must be greater than %d."):format(_Info[self].Version or 0), 2)
					end
				end,
				Type = Number,
			}

			-- _Enabled
			property "_Enabled" {
				Get = function(self)
					return IsEnabled(self)
				end,
				Set = function(self, enabled)
					if enabled then
						Enable(self)
					else
						Disable(self)
					end
				end,
				Type = Boolean,
			}

			-- _Parent
			property "_Parent" {
				Get = function(self)
					return _Info[self].Parent
				end,
			}
		end

		------------------------------------------------------
		-- Script Handler
		------------------------------------------------------
		local function OnEvent(self, event, ...)
			if type(self[event]) == "function" then
				return self[event](self, ...)
			end
		end

		------------------------------------------------------
		-- Constructor
		------------------------------------------------------
		function Module(parent, name)
			if not Object.IsClass(parent, Module) and not Object.IsClass(parent, Addon) then
				error("Usage : Module(mdl, name) : 'mdl' - Addon or Module expected.", 2)
			end

			if type(name) ~= "string"  then
				error(("Usage : Module(mdl, name) : 'name' - string expected, got %s."):format(type(name)), 2)
			end

			name = name:match("[_%w]+")

			if not name or name == "" then return end

			_Info[parent].Module = _Info[parent].Module or {}
			_Info[parent].Module[name] = Object()

			local mdl = _Info[parent].Module[name]

			rawset(parent, name, mdl)

			_Info[mdl] = {
				Owner = mdl,
				Name = name,
				Parent = parent,
			}

			_Info[mdl].DefaultState = true
			_Info[mdl].Disabled = _Info[parent].Disabled

			mdl.OnEvent = mdl.OnEvent + OnEvent

			return mdl
		end

		------------------------------------------------------
		-- Exist checking
		------------------------------------------------------
		function __exist(cls, parent, name)
			if (Object.IsClass(parent, Module) or Object.IsClass(parent, Addon)) and type(name) == "string" then
				name = name:match("[_%w]+")

				return name and _Info[parent].Module and _Info[parent].Module[name]
			end
		end

		------------------------------------------------------
		-- __index for class instance
		------------------------------------------------------
		function __index(self, key)
			if _Special_KeyWord[key] then
				return _Special_KeyWord[key]
			end

			if _Class_KeyWords[key] then
				return _Class_KeyWords[key]
			end

			-- Check namespace
			local ns = rawget(self, "__IGAS_NameSpace")
			local parent = _Info[self].Parent

			while not ns and parent do
				ns = rawget(parent, "__IGAS_NameSpace")
				parent = _Info[parent].Parent
			end

			if IGAS:GetClassInfo(ns) then
				if key == IGAS:GetClassInfo(ns) then
					self[key] = ns
					return rawget(self, key)
				elseif ns[key] then
					self[key] = ns[key]
					return rawget(self, key)
				end
			end

			-- Check imports
			if _Info[self].Import then
				for _, ns in ipairs(_Info[self].Import) do
					if key == IGAS:GetClassInfo(ns) then
						self[key] = ns
						return rawget(self, key)
					elseif ns[key] then
						self[key] = ns[key]
						return rawget(self, key)
					end
				end
			end

			return _Info[self].Parent and _Info[self].Parent[key]
		end

		------------------------------------------------------
		-- __newindex for class instance
		------------------------------------------------------
		function __newindex(self, key, value)
			if _Special_KeyWord[key] or _Class_KeyWords[key] then
				error(("%s is a keyword."):format(key), 2)
			end

			rawset(self, key, value)
		end
	endclass "Module"

	------------------------------------------------------
	-- Script
	------------------------------------------------------
	do
		script "OnLoad"
		script "OnEnable"
		script "OnDisable"
		script "OnDispose"
		script "OnSlashCmd"
	end

	------------------------------------------------------
	-- Method
	------------------------------------------------------
	do
		if _EventManager then
			------------------------------------
			--- Register a event
			-- @name Addon:RegisterEvent
			-- @class function
			-- @param event the event's name
			-- @return nil
			-- @Usage Addon:RegisterEvent("CUSTOM_EVENT_1")
			------------------------------------
			function RegisterEvent(self, event)
				if type(event) == "string" and event ~= "" then
					return _EventManager:Register(event, self)
				else
					error(("Usage : Addon:RegisterEvent(event) : 'event' - string expected, got %s."):format(type(event) == "string" and "empty string" or type(event)), 2)
				end
			end

			------------------------------------
			--- Check if  a event is registered for an Addon
			-- @name Addon:IsEventRegistered
			-- @class function
			-- @param event the event's name
			-- @return true if the event is registered to that Addon
			-- @Usage Addon:IsEventRegistered("CUSTOM_EVENT_1")
			------------------------------------
			function IsEventRegistered(self, event)
				return _EventManager:IsRegistered(event, self)
			end

			------------------------------------
			--- Undo register all events for the Addon
			-- @name Addon:UnregisterAllEvents
			-- @class function
			-- @return nil
			-- @Usage Addon:UnregisterAllEvents()
			------------------------------------
			function UnregisterAllEvents(self)
				return _EventManager:UnregisterAll(self)
			end

			------------------------------------
			--- Undo Register a event for an Addon
			-- @name Addon:UnregisterEvent
			-- @class function
			-- @param event the event's name
			-- @return nil
			-- @Usage Addon:UnregisterEvent("CUSTOM_EVENT_1")
			------------------------------------
			function UnregisterEvent(self, event)
				return _EventManager:Unregister(event, self)
			end
		end

		------------------------------------------------------
		-- SlashCmd System
		------------------------------------------------------
		if _G["SlashCmdList"] then
			_SlashFuncs = _SlashFuncs or {}

			-- SlashCmd Operation
			local function parsing(self, ...)
				Object.Fire(self, "OnSlashCmd", ...)

				if _Info[self].Module then
					for _, mdl in pairs(_Info[self].Module) do
						parsing(mdl, ...)
					end
				end
			end

			local function GetSlashCmdArgs(msg, input)
				local option, info

				if msg and type(msg) == "string" and msg ~= "" then
					option, info = msg:match("(%a+)%s*(.*)")

					if option == "" then option = nil end
					if info == "" then info = nil end
				end

				if IGAS.GetWrapper then
					-- Wrapper the inputbox if IGAS GUI lib is loaded.
					input = IGAS:GetWrapper(input)
				end

				return option, info, input
			end

			local function GetSlashCmd(name)
				_SlashFuncs[name] = _SlashFuncs[name] or function(msg, input)
					if _Addon[name] then
						parsing(_Addon[name], GetSlashCmdArgs(msg, input))
					end
				end
				return _SlashFuncs[name]
			end

			------------------------------------
			--- Add slash commands to an addon
			-- @name Addon:AddSlashCmd
			-- @class function
			-- @param slash1 the slash commands, a string started with "/", if not started with "/", will auto add it.
			-- @param slash2, ... Optional,other slash commands
			-- @return nil
			-- @usage Addon:AddSlashCmd("/hw", hw2")
			------------------------------------
			function AddSlashCmd(self, ...)
				if select('#', ...) == 0 then
					return nil
				end

				local AddName = _Info[self].Name

				local slash = {}
				local sl
				for i = 1, select('#', ...) do
					sl = select(i, ...)
					if sl and type(sl) == "string" then
						if sl:match("^/%S+") then
							slash[sl:match("^/%S+")] = true
						elseif sl:match("^%S+") then
							slash["/"..sl:match("^%S+")] = true
						end
					end
				end

				if not next(slash) then
					error("Usage : Addon:AddSlashCmd(slash1 [, slash2, ...]) : no slash commands found.", 2)
				end

				local index = 1

				while _G["SLASH_"..AddName..index] do
					slash[_G["SLASH_"..AddName..index]] = nil
					index = index + 1
				end

				for cmd in pairs(slash) do
					_G["SLASH_"..AddName..index] = cmd
					index = index + 1
				end

				if not _G["SlashCmdList"][AddName] then
					_G["SlashCmdList"][AddName] = GetSlashCmd(AddName)
				end
			end
		end

		------------------------------------
		--- Add SaveVariables to the addon
		-- @name Addon:AddSavedVariable
		-- @class function
		-- @param name the savevariable's name
		-- @return the savevariable, no need to receive it, the name passed in this function will be the identify.
		-- @usage Addon:AddSavedVariable("HelloWorld_SaveV")
		------------------------------------
		function AddSavedVariable(self, name)
			if type(name) ~= "string" or strtrim(name) == "" then
				error(("Usage : Addon:AddSavedVariable(name) : 'name' - string expected, got %s."):format(type(name) == "string" and "empty string" or type(name)), 2)
			end

			name = strtrim(name)
			if not _Logined then
				_Info[self].SavedVariables = _Info[self].SavedVariables or {}
				_Info[self].SavedVariables[name] = true
			end

			_G[name] = _G[name] or {}
			self[name] = _G[name]

			return self[name]
		end

		------------------------------------
		--- Get the addon's information by the given name.
		-- @name Addon:GetMetadata
		-- @class function
		-- @param field the informations' name
		-- @return infomations stored in the addon
		-- @usage Addon:GetMetadata("Author")    -- return "Kurapica"
		------------------------------------
		function GetMetadata(self, field)
			if type(field) ~= "string" or strtrim(field) == "" then
				error(("Usage : Addon:GetMetadata(field) : 'field' - string expected, got %s."):format(type(field) == "string" and "empty string" or type(field)), 2)
			end

			return self._Metadata[field]
		end

		------------------------------------
		--- Set the addon's information by the given name.
		-- @name Addon:SetMetadata
		-- @class function
		-- @param field the information's name
		-- @param value the information's value
		-- @usage Addon:SetMetadata("Author", "Kurapica")
		------------------------------------
		function SetMetadata(self, field, value)
			if type(field) ~= "string" or strtrim(field) == "" then
				error(("Usage : Addon:SetMetadata(field, value) : 'field' - string expected, got %s."):format(type(field) == "string" and "empty string" or type(field)), 2)
			end

			self._Metadata[field] = value
		end

		------------------------------------
		--- Enable the Addon.
		-- @name Addon:Enable
		-- @class function
		-- @return nil
		-- @usage Addon:Enable()
		------------------------------------
		function Enable(self)
			if _Info[self].Disabled then
				_Info[self].Disabled= nil

				if _Logined then Object.Fire(self, "OnEnable") end

				if not _Info[self].Module then return end

				for _, mdl in pairs(_Info[self].Module) do
					if _Info[mdl].DefaultState ~= false then
						Module.Enable(mdl)
					end
				end
			end
		end

		------------------------------------
		--- Disable the Addon.The Addon won't receive events.
		-- @name Addon:Disable
		-- @class function
		-- @return nil
		-- @usage Addon:Disable()
		------------------------------------
		function Disable(self)
			if not _Info[self].Disabled then
				_Info[self].Disabled = true

				if _Logined then Object.Fire(self, "OnDisable") end

				if not _Info[self].Module then return end

				for _, mdl in pairs(_Info[self].Module) do
					_Info[mdl].DefaultState = not _Info[mdl].Disabled
					if not _Info[mdl].Disabled then
						Module.Disable(mdl)
					end
				end
			end
		end

		------------------------------------
		--- Check if the Addon is enabled
		-- @name Addon:IsEnabled
		-- @class function
		-- @return true if the Addon is enabled
		-- @usage Addon:IsEnabled()
		------------------------------------
		function IsEnabled(self)
			return not _Info[self].Disabled
		end

		------------------------------------
		--- Add a frame to Addon.
		-- @name Addon:AddFrame
		-- @class function
		-- @param frame the user-defined frame that needed to be add to the Addon, must be the parent of other user-defined frames, will be auto-diposed when Module is dispoing.Also make the frame can be access in the addon's scope by it's name.
		-- @return nil
		-- @usage Addon:AddFrame(MyFrame1)
		------------------------------------
		function AddFrame(self, frame)
			if type(frame) == "table" and frame.Name then
				_Info[self].Frames = _Info[self].Frames or setmetatable({}, _MetaWK)
				_Info[self].Frames[frame] = true
				self[frame.Name] = frame
			end
		end

		------------------------------------
		--- Remove a frame from Addon.
		-- @name Addon:RemoveFrame
		-- @class function
		-- @param frame the user-defined frame that needed to be removed from the Addon
		-- @return nil
		-- @usage Addon:RemoveFrame(MyFrame1)
		------------------------------------
		function RemoveFrame(self, frame)
			if frame and _Info[self].Frames[frame] then
				_Info[self].Frames[frame] = nil
				if self[frame.Name] == frame then
					self[frame.Name] = nil
				end
			end
		end

		------------------------------------
		--- This is using to dispose a Addon, and set the enviroment to it's container if needed.
		-- @name Addon:Dispose
		-- @class function
		-- @return nil
		-- @usage Addon:Dispose()
		------------------------------------
		function Dispose(self)
			local info = _Info[self]
			local chk, ret

			Object.Fire(self, "OnDispose")

			if info.Frames then
				for fr in pairs(info.Frames) do
					if type(fr) == "table" and type(fr.Dispose) == "function" then
						chk, ret = pcall(fr.Dispose, fr)

						if not chk then
							errorhandler(ret)
						end
					end
				end
			end

			self:UnregisterAllEvents()

			if info.Module then
				for _, mdl in pairs(info.Module) do
					chk, ret = pcall(Module.Dispose, mdl)

					if not chk then
						errorhandler(ret)
					end
				end
			end

			_Addon[info.Name] = nil

			if getfenv(2) == self then
				setfenv(2, _G)
			end

			wipe(info)
			_Info[self] = nil

			return Object.Dispose(self)
		end

		------------------------------------
		--- Create or get a child-module of this Addon.And set the environment to the child-module
		-- @name Addon:NewModule
		-- @class function
		-- @param name the child-module's name,can be "Mdl1.SubMdl1.SubSubMdl1", use dot to concat
		-- @param version Optional,used for version control, if there is another version, and the old version is equal or great than this one, return nil, else return the child-module
		-- @return the child-module
		-- @usage Addon:NewModule("Mdl_MyFrame1")
		------------------------------------
		function NewModule(self, name, version)
			if type(name) ~= "string" or strtrim(name) == "" then
				error(("Usage : Addon:NewModule(name, [version]) : 'name' - string expected, got %s."):format(type(name) == "string" and "empty string" or type(name)), 2)
			end

			if version then
				if type(version) ~= "number" then
					error(("Usage : Addon:NewModule(name, [version]) : 'version' - number expected, got %s."):format(type(version)), 2)
				elseif version < 1 then
					error("Usage : Addon:NewModule(name, [version]) : 'version' - must be greater than 0.", 2)
				end
			end

			local module = self

			for sub in name:gmatch("[^%.]+") do
				sub = sub and strtrim(sub)
				if not sub or sub =="" then return end

				module = Module(module, sub)

				if not module then return end
			end

			if not module or module == self then return end

			if version and _Info[module].Version and _Info[module].Version >= version then
				return
			elseif version then
				_Info[module].Version = version
			end

			setfenv(2, module)

			return module
		end

		------------------------------------
		--- Get the child-module of this Addon.
		-- @name Addon:GetModule
		-- @class function
		-- @param name the child-module's name,can be "Mdl1.SubMdl1.SubSubMdl1", use dot to concat
		-- @param create if true, the module will be created if not existed
		-- @return the child-module
		-- @usage Addon:GetModule("Mdl_MyFrame1")
		------------------------------------
		function GetModule(self, name, create)
			if type(name) ~= "string" or strtrim(name) == "" then
				error(("Usage : Addon:GetModule(name, [create]) : 'name' - string expected, got %s."):format(type(name) == "string" and "empty string" or type(name)), 2)
			end

			local module = self

			if create then
				for sub in name:gmatch("[^%.]+") do
					sub = sub and strtrim(sub)
					if not sub or sub =="" then return end

					module = Module(module, sub)

					if not module then return end
				end
			else
				for sub in name:gmatch("[^%.]+") do
					sub = sub and strtrim(sub)
					if not sub or sub =="" then return end

					module =  _Info[module].Module and _Info[module].Module[sub]

					if not module then return end
				end
			end

			if module == self then return end

			return module
		end
	end

	------------------------------------------------------
	-- Special keywords
	------------------------------------------------------
	_Special_KeyWord = _Special_KeyWord or {}
	do
		------------------------------------
		--- import classes from the given name's namespace to the current Addon
		-- @name import
		-- @class function
		-- @param name the namespace's name list, using "." to split.
		-- @usage import "System.Widget"
		------------------------------------
		function _Special_KeyWord.import(name)
			if type(name) ~= "string" then
				error([[Usage: import "namespaceA.namespaceB"]], 2)
			end

			if type(name) == "string" and name:find("%.%s*%.") then
				error("the namespace 's name can't have empty string between dots.", 2)
			end

			local env = getfenv(2)

			local info = _Info[env]

			if not info then
				error("can't use import here.", 2)
			end

			local ns = IGAS:GetNameSpace(name)

			if not ns then
				error(("no namespace is found with name : %s"):format(name), 2)
			end

			info.Import = info.Import or {}

			for _, v in ipairs(info.Import) do
				if v == ns then
					return
				end
			end

			tinsert(info.Import, ns)
		end
	end

	------------------------------------------------------
	-- Property
	------------------------------------------------------
	do
		-- _Addon
		property "_Addon" {
			Get = function(self)
				return self
			end,
		}

		-- _Name
		property "_Name" {
			Get = function(self)
				return _Info[self].Name
			end,
		}

		-- _Version
		property "_Version" {
			Get = function(self)
				return _Info[self].Version or 0
			end,
			Set = function(self, version)
				if version > (_Info[self].Version or 0) then
					_Info[self].Version = version
				else
					error(("The version must be greater than %d."):format(_Info[self].Version or 0), 2)
				end
			end,
			Type = Number,
		}

		-- _Enabled
		property "_Enabled" {
			Get = function(self)
				return IsEnabled(self)
			end,
			Set = function(self, enabled)
				if enabled then
					Enable(self)
				else
					Disable(self)
				end
			end,
			Type = Boolean,
		}

		-- _AutoWrapper
		property "_AutoWrapper" {
			Get = function(self)
				return not _Info[self].NoAutoWrapper
			end,
			Set = function(self, auto)
				_Info[self].NoAutoWrapper = not auto
			end,
			Type = Boolean,
		}

		-- _Metadata
		property "_Metadata" {
			Get = function(self)
				_Info[self].MetaData = _Info[self].MetaData or setmetatable({}, {
					__index = function(self, key)
						local value = GetAddOnMetadata and GetAddOnMetadata(_Info[self].Name, key)

						if value ~= nil then
							self[key] = value
						end

						return rawget(self, key)
					end,
					__newindex = function(self, key, value)
						if type(key) == "string" and strtrim(key) ~= "" then
							rawset(self, strtrim(key), value)
						end
					end,
				})

				return _Info[self].MetaData
			end,
		}
	end

	------------------------------------------------------
	-- Script Handler
	------------------------------------------------------
	local function OnEvent(self, event, ...)
		if type(self[event]) == "function" then
			return self[event](self, ...)
		end
	end

	------------------------------------------------------
	-- Constructor
	------------------------------------------------------
	function Addon(name)
		if type(name) ~= "string" then
			error(("Usage : Addon(name) : 'name' - string expected, got %s."):format(type(name)), 2)
		end

		name = name:match("[_%w]+")

		if not name or name == "" then return end

		_Addon[name] = Object()

		local addon = _Addon[name]

		_Info[addon] = {
			Owner = addon,
			Name = name,
		}

		addon.OnEvent = addon.OnEvent + OnEvent

		return addon
	end

	------------------------------------------------------
	-- Exist checking
	------------------------------------------------------
	function __exist(cls, name)
		if type(name) ~= "string" then
			return
		end

		name = name:match("[_%w]+")

		return name and _Addon[name]
	end

	------------------------------------------------------
	-- __index for class instance
	------------------------------------------------------
	function __index(self, key)
		if _Special_KeyWord[key] then
			return _Special_KeyWord[key]
		end

		if _Class_KeyWords[key] then
			return _Class_KeyWords[key]
		end

		-- Check namespace
		local ns = rawget(self, "__IGAS_NameSpace")
		if IGAS:GetClassInfo(ns) then
			if key == IGAS:GetClassInfo(ns) then
				self[key] = ns
				return rawget(self, key)
			elseif ns[key] then
				self[key] =ns[key]
				return rawget(self, key)
			end
		end

		-- Check imports
		if _Info[self].Import then
			for _, ns in ipairs(_Info[self].Import) do
				if key == IGAS:GetClassInfo(ns) then
					self[key] = ns
					return rawget(self, key)
				elseif ns[key] then
					self[key] = ns[key]
					return rawget(self, key)
				end
			end
		end

		if key ~= "_G" and type(key) == "string" and key:find("^_") then
			return
		end

		if _G[key] then
			if type(_G[key]) == "function" or type(_G[key]) == "table" then
				if not _Info[self].NoAutoWrapper and IGAS.GetWrapper then
					rawset(self, key, IGAS:GetWrapper(_G[key]))
				else
					rawset(self, key, _G[key])
				end
				return rawget(self, key)
			end

			return _G[key]
		end
	end

	------------------------------------------------------
	-- __newindex for class instance
	------------------------------------------------------
	function __newindex(self, key, value)
		if _Special_KeyWord[key] or _Class_KeyWords[key] then
			error(("%s is a keyword."):format(key), 2)
		end

		rawset(self, key, value)
	end
endclass "Addon"

------------------------------------------------------
-- Global Settings
------------------------------------------------------
do
	------------------------------------
	--- Create or Get an addon by name, then change the environment to the addon if not set setEnv to false
	-- @name IGAS:NewAddon
	-- @class function
	-- @param addonName	the addon's name, must be unique, also can be a serie tokens seperate by ".",the first token is addon' name, the second token will be a child-module to the addon, the third token will create the child-module to the second, and go on.
	-- @param version Optional,used for version control, if there is another version, and the old version is equal or great than this one, return nil, else return the child-module
	-- @param info Optional,Optional,must be a table, contains informations such as "author", "version" .etc
	-- @return addon that a table contains the addon's scope, no need to receive it if you don't set setEnv to false, after call this function, you can using "_Addon" to access the addon
	-- @usage local addon = IGAS:NewAddon("HelloWorld", 1, {Author="Kurapica", Version="1.0", Note="Hello World Example", Interface="30010", Title="Hello World"})
	-- @usage IGAS:NewAddon("HelloWorld")
	------------------------------------
	function IGAS:NewAddon(name, version, info)
		if type(name) ~= "string" or strtrim(name) == "" then
			error(("Usage : IGAS:NewAddon(name, [version]) : 'name' - string expected, got %s."):format(type(name) == "string" and "empty string" or type(name)), 2)
		end

		if version then
			if type(version) ~= "number" then
				error(("Usage : IGAS:NewAddon(name, [version]) : 'version' - number expected, got %s."):format(type(version)), 2)
			elseif version < 1 then
				error("Usage : IGAS:NewAddon(name, [version]) : 'version' - must be greater than 0.", 2)
			end
		end

		local module = _Addon
		local addon

		for sub in name:gmatch("[^%.]+") do
			sub = sub and strtrim(sub)
			if not sub or sub =="" then return end

			if module == _Addon then
				module = Addon(sub)
				addon = module
			else
				module = Addon.Module(module, sub)
			end

			if not module then return end
		end

		if not module or module == _Addon then return end

		if version and module._Version >= version then
			return
		elseif version then
			 module._Version = version
		end

		-- MetaData
		if type(info) == "table" and next(info) then
			for field, v in pairs(info) do
				if type(field) == "string" and strtrim(field) ~= "" then
					addon._Metadata[field] = v
				end
			end
		end

		setfenv(2, module)

		return module
	end

	------------------------------------
	--- Get an addon by name
	-- @name IGAS:GetAddon
	-- @class function
	-- @param addonName	the addon's name, must be unique, also can be a serie tokens seperate by ".",the first token is addon' name, the second token will be a child-module to the addon, the third token will create the child-module to the second, and go on.
	-- @return addon that a table contains the addon's scope, no need to receive it if you don't set setEnv to false, after call this function, you can using "_Addon" to access the addon
	-- @usage local addon = IGAS:GetAddon("HelloWorld")
	-- @usage IGAS:GetAddon("HelloWorld")
	------------------------------------
	function IGAS:GetAddon(name)
		if type(name) ~= "string" or strtrim(name) == "" then
			error(("Usage : IGAS:GetAddon(name) : 'name' - string expected, got %s."):format(type(name) == "string" and "empty string" or type(name)), 2)
		end

		local module = _Addon

		for sub in name:gmatch("[^%.]+") do
			sub = sub and strtrim(sub)
			if not sub or sub =="" then return end

			if module == _Addon then
				module = _Addon[sub]
			else
				module = module:GetModule(sub)
			end

			if not module then return end
		end

		if not module or module == _Addon then return end

		return module
	end

	------------------------------------
	--- Get a list of the existed addons' name
	-- @name IGAS:GetAddonList
	-- @class function
	-- @return list a table contains all the existed addons' name
	-- @usage IGAS:GetAddonList()
	------------------------------------
	function IGAS:GetAddonList()
		local lst = {}

		for name in pairs(_Addon) do
			tinsert(lst, name)
		end

		return lst
	end
end
